from bsb.morphologies import Morphology, Branch, _Labels
from bsb.exceptions import MorphologyRepositoryError, MissingMorphologyError
from bsb.storage.interfaces import (
    MorphologyRepository as IMorphologyRepository,
    StoredMorphology,
)
from .resource import Resource
import numpy as np
import json
import itertools

_root = "/morphologies"


class MorphologyRepository(Resource, IMorphologyRepository):
    def __init__(self, engine):
        super().__init__(engine, _root)

    def select(self, *selectors):
        all_loaders = self.all()
        selected = []
        for selector in selectors:
            selector.validate(all_loaders)
            selected.extend(filter(selector.pick, all_loaders))
        return selected

    def preload(self, name):
        return StoredMorphology(name, self._make_loader(name), self.get_meta(name))

    def _make_loader(self, name):
        def loader():
            return self.load(name)

        return loader

    def get_meta(self, name):
        with self._engine._read():
            with self._engine._handle("r") as repo:
                try:
                    meta = _meta(repo[f"{self._path}/{name}/"])
                except KeyError:
                    raise MissingMorphologyError(
                        f"`{self._engine.root}` contains no morphology named `{name}`."
                    ) from None
        meta["name"] = name
        return meta

    def all(self):
        return [self.preload(name) for name in self.keys()]

    def has(self, name):
        with self._engine._read():
            with self._engine._handle("r") as repo:
                return f"{self._path}/{name}" in repo

    def load(self, name):
        with self._engine._read():
            with self._engine._handle("r") as repo:
                try:
                    root = repo[f"{self._path}/{name}/"]
                except Exception:
                    raise MissingMorphologyError(
                        f"`{self._engine.root}` contains no morphology named `{name}`."
                    ) from None
                data = root["data"][()]
                points = data[:, :3].copy()
                radii = data[:, 3].copy()
                # Turns the forced JSON str keys back into ints
                labelsets = {
                    int(k): v for k, v in json.loads(root["data"].attrs["labels"]).items()
                }
                labels = _Labels(
                    len(points), buffer=data[:, 4].astype(int), labels=labelsets
                )
                prop_names = root["data"].attrs["properties"]
                props = dict(zip(prop_names, np.rollaxis(data[:, 5:], 1)))
                parents = {-1: None}
                branch_id = itertools.count()
                roots = []
                ptr = 0
                for nptr, p in root["graph"][()]:
                    radii[ptr:nptr]
                    labels[ptr:nptr]
                    {k: v[ptr:nptr] for k, v in props.items()}
                    branch = Branch(
                        points[ptr:nptr],
                        radii[ptr:nptr],
                        labels[ptr:nptr],
                        {k: v[ptr:nptr] for k, v in props.items()},
                    )
                    parent = parents.get(p, None)
                    parents[next(branch_id)] = branch
                    if parent:
                        parent.attach_child(branch)
                    else:
                        roots.append(branch)
                    ptr = nptr
                meta = _meta(root)
                meta["name"] = name
                morpho = Morphology(
                    roots, meta, shared_buffers=(points, radii, labels, props)
                )
                assert morpho._check_shared(), "Morpho read with unshareable buffers"
                return morpho

    def save(self, name, morphology, overwrite=False):
        with self._engine._write():
            with self._engine._handle("a") as repo:
                me = repo[self._path]
                if self.has(name):
                    if overwrite:
                        self.remove(name)
                    else:
                        root = self._engine.root
                        raise MorphologyRepositoryError(
                            f"A morphology called '{name}' already exists in `{root}`."
                        )
                root = me.create_group(name)
                # Optimizing a morphology goes through the same steps as what is required
                # to save it to disk; plus, now the user's object is optimized :)
                morphology.optimize()
                branches = morphology.branches
                n_prop = len(morphology._shared._prop)
                data = np.empty((len(morphology), 5 + n_prop))
                data[:, :3] = morphology._shared._points
                data[:, 3] = morphology._shared._radii
                data[:, 4] = morphology._shared._labels
                for i, prop in enumerate(morphology._shared._prop.values()):
                    data[:, 5 + i] = prop
                dds = root.create_dataset("data", data=data)
                dds.attrs["labels"] = json.dumps(
                    {k: list(v) for k, v in morphology._shared._labels.labels.items()}
                )
                dds.attrs["properties"] = [*morphology._shared._prop.keys()]
                graph = np.empty((len(branches), 2))
                parents = {None: -1}
                ptr = 0
                for i, branch in enumerate(morphology.branches):
                    ptr += len(branch)
                    graph[i, 0] = ptr
                    graph[i, 1] = parents[branch.parent]
                    parents[branch] = i
                root.create_dataset("graph", data=graph, dtype=int)
                try:
                    for k, v in morphology.meta.items():
                        root.attrs[f"meta:{k}"] = v if v is not None else np.nan
                except Exception:
                    raise MorphologyRepositoryError(
                        f"Trying to store invalid {type(v)} metadata '{k}' on `{name}`."
                    ) from None
                root.attrs["meta:ldc"] = np.min(morphology._shared._points, axis=0)
                root.attrs["meta:mdc"] = np.max(morphology._shared._points, axis=0)

    def remove(self, name):
        with self._engine._write():
            with self._engine._handle("a") as repo:
                try:
                    del repo[f"{self._path}/{name}"]
                except KeyError:
                    raise MorphologyRepositoryError(f"'{name}' doesn't exist.") from None


def _meta(group):
    return dict((k[5:], v) for k, v in group.attrs.items() if k.startswith("meta:"))



from bsb.exceptions import DatasetNotFoundError
from .resource import Resource
from bsb.storage._chunks import Chunk
from bsb.storage.interfaces import ConnectivitySet as IConnectivitySet
import numpy as np

_root = "/connectivity/"


class ConnectivitySet(Resource, IConnectivitySet):
    """
    Fetches placement data from storage.

    .. note::

        Use :meth:`Scaffold.get_connectivity_set <bsb.core.Scaffold.get_connectivity_set>`
        to correctly obtain a :class:`~bsb.storage.interfaces.ConnectivitySet`.
    """

    def __init__(self, engine, tag):
        self.tag = tag
        self.pre = None
        self.post = None
        super().__init__(engine, _root + tag)
        with engine._read():
            with engine._handle("r") as h:
                if not self.exists(engine, tag, handle=h):
                    raise DatasetNotFoundError(f"ConnectivitySet '{tag}' does not exist")
                self._pre_name = h[self._path].attrs["pre"]
                self._post_name = h[self._path].attrs["post"]

    def __len__(self):
        return sum(len(data[0]) for _, _, _, data in self.flat_iter_connections("inc"))

    @classmethod
    def get_tags(cls, engine):
        """
        Returns all the connectivity tags in the network.
        """
        with engine._read():
            with engine._handle("r") as h:
                return list(h[_root].keys())

    @classmethod
    def create(cls, engine, pre_type, post_type, tag=None):
        """
        Create the structure for this connectivity set in the HDF5 file. Connectivity sets
        are stored under ``/connectivity/<tag>``.
        """
        if tag is None:
            tag = f"{pre_type.name}_to_{post_type.name}"
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                g = h.create_group(path)
                g.attrs["pre"] = pre_type.name
                g.attrs["post"] = post_type.name
                g.require_group(path + "/inc")
                g.require_group(path + "/out")
        cs = cls(engine, tag)
        cs.pre = pre_type
        cs.post = post_type
        return cs

    @staticmethod
    def exists(engine, tag, handle=None):
        """
        Checks whether a :class:`~.connectivity_set.ConnectivitySet` with the given tag
        exists.

        :param engine: Engine to use for the lookup.
        :type engine: :class:`.HDF5Engine`
        :param tag: Tag of the set to look for.
        :type tag: str
        :param handle: An open handle to use instead of opening one.
        :type handle: :class:`h5py.File`
        :returns: Whether the tag exists.
        :rtype: bool
        """

        def check(h):
            return _root + tag in h

        if handle is not None:
            return check(handle)
        else:
            with engine._read():
                with engine._handle("r") as h:
                    return check(h)

    @classmethod
    def require(cls, engine, pre_type, post_type, tag=None):
        """
        Get or create a :class:`~.connectivity_set.ConnectivitySet`.

        :param engine: Engine to fetch/write the data.
        :type engine: :class:`.HDF5Engine`
        :param pre_type: Presynaptic cell type.
        :type pre_type: :class:`~bsb.cell_types.CellType`
        :param post_type: Postsynaptic cell type.
        :type post_type: :class:`~bsb.cell_types.CellType`
        :param tag: Tag to store the set under. Defaults to
          ``{pre_type.name}_to_{post_type.name}``.
        :type tag: str
        :returns: Existing or new connectivity set.
        :rtype: :class:`~.connectivity_set.ConnectivitySet`
        """
        if tag is None:
            tag = f"{pre_type.name}_to_{post_type.name}"
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                g = h.require_group(path)
                if g.attrs.setdefault("pre", pre_type.name) != pre_type.name:
                    raise ValueError(
                        "Given and stored type mismatch:"
                        + f" {pre_type.name} vs {g.attrs['pre']}"
                    )
                if g.attrs.setdefault("post", post_type.name) != post_type.name:
                    raise ValueError(
                        "Given and stored type mismatch:"
                        + f" {post_type.name} vs {g.attrs['post']}"
                    )
                g.require_group(path + "/inc")
                g.require_group(path + "/out")
        cs = cls(engine, tag)
        cs.pre = pre_type
        cs.post = post_type
        return cs

    def clear(self):
        raise NotImplementedError("Will do once I have some sample data :)")

    def connect(self, pre_set, post_set, src_locs, dest_locs):
        src_locs = np.array(src_locs, copy=False)
        dest_locs = np.array(dest_locs, copy=False)
        if not len(src_locs):
            return
        if len(src_locs) != len(dest_locs):
            raise ValueError("Location matrices must be of same length.")
        with self._engine._write():
            with self._engine._handle("a") as handle:
                for data in self._demux(pre_set, post_set, src_locs, dest_locs):
                    if not len(data[-1]):
                        # Don't write empty data
                        continue
                    self.chunk_connect(*data, handle=handle)

    def _demux(self, pre, post, src_locs, dst_locs):
        dst_chunks = post.get_loaded_chunks()
        lns = []
        for dst in iter(dst_chunks):
            with post.chunk_context(dst):
                lns.append(len(post))
        # Iterate over each source chunk
        for src in pre.get_loaded_chunks():
            # Count the number of cells
            with pre.chunk_context(src):
                ln = len(pre)
            src_idx = src_locs[:, 0] < ln
            src_block = src_locs[src_idx]
            dst_block = dst_locs[src_idx]
            if len(dst_chunks) == 1:
                block_idx = np.lexsort((dst_block[:, 0], src_block[:, 0]))
                yield src, dst, src_block[block_idx], dst_block[block_idx]
            else:
                dctr = 0
                for dst, dln in zip(iter(dst_chunks), lns):
                    block_idx = (dst_block[:, 0] >= dctr) & (dst_block[:, 0] < dctr + dln)
                    yield src, dst, src_block[block_idx], dst_block[block_idx]
                    dctr += dln
            src_locs = src_locs[~src_idx]
            dst_locs = dst_locs[~src_idx]
            # We sifted `ln` cells out of the dataset, so reduce the ids.
            src_locs[:, 0] -= ln

    def _store_pointers(self, group, chunk, n, total):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs.get("chunk_list", [])]
        if chunk in chunks:
            # Source chunk already existed, just increment the subseq. pointers
            inc_from = chunks.index(chunk) + 1
        else:
            # We are the last chunk, we start adding rows at the end.
            group.attrs[str(chunk.id)] = total
            # Move up the increment pointer to place ourselves after the
            # last element
            chunks.append(chunk)
            inc_from = len(chunks)
            group.attrs["chunk_list"] = chunks
        # Increment the pointers of the chunks that follow us, by `n` rows.
        for c in chunks[inc_from:]:
            group.attrs[str(c.id)] += n

    def _get_sorted_pointers(self, group):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs["chunk_list"]]
        ptrs = np.array([group.attrs[str(c.id)] for c in chunks])
        sorted = np.argsort(ptrs)
        chunks = [chunks[cid] for cid in sorted]
        return chunks, ptrs[sorted]

    def _get_insert_pointers(self, group, chunk):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs["chunk_list"]]
        iptr = group.attrs[str(chunk.id)]
        idx = chunks.index(chunk)
        if idx + 1 == len(chunks):
            # Last chunk, no end pointer
            eptr = None
        else:
            # Get the pointer of the next chunk
            eptr = group.attrs[str(chunks[idx + 1].id)]
        return iptr, eptr

    def _get_chunk_data(self, dest_chunk):
        with self._engine._write():
            with self._engine._handle("a") as h:
                grp = h[f"{self._path}/{dest_chunk.id}"]
                src_chunks = grp.attrs["chunk_list"]
                chunk_ptrs = [grp.attrs[str(Chunk(c, (0, 0, 0)).id)] for c in src_chunks]
                src = grp["global_locs"][()]
                dest = grp["local_locs"][()]
        return src_chunks, chunk_ptrs, src, dest

    def chunk_connect(self, src_chunk, dst_chunk, src_locs, dst_locs, handle=None):
        if len(src_locs) != len(dst_locs):
            raise ValueError("Location matrices must be of same length.")
        if handle is None:
            with self._engine._write():
                with self._engine._handle("a") as handle:
                    self._connect(src_chunk, dst_chunk, src_locs, dst_locs, handle)
        else:
            self._connect(src_chunk, dst_chunk, src_locs, dst_locs, handle)

    def _connect(self, src_chunk, dest_chunk, lloc, gloc, handle):
        self._insert("inc", dest_chunk, src_chunk, gloc, lloc, handle)
        self._insert("out", src_chunk, dest_chunk, lloc, gloc, handle)

    def _insert(self, tag, local_, global_, lloc, gloc, handle):
        grp = handle.require_group(f"{self._path}/{tag}/{local_.id}")
        unpack_me = [None, None]
        # require_dataset doesn't work for resizable datasets, see
        # https://github.com/h5py/h5py/issues/2018
        # So we create a little thingy for requiring src & dest
        for i, tag in enumerate(("global_locs", "local_locs")):
            if tag in grp:
                unpack_me[i] = grp[tag]
            else:
                unpack_me[i] = grp.create_dataset(
                    tag, shape=(0, 3), dtype=int, chunks=(1024, 3), maxshape=(None, 3)
                )
        lcl_ds, gbl_ds = unpack_me
        # Move the pointers that keep track of the chunks
        new_rows = len(lloc)
        total = len(lcl_ds)
        self._store_pointers(grp, global_, new_rows, total)
        iptr, eptr = self._get_insert_pointers(grp, global_)
        if eptr is None:
            eptr = total + new_rows
        # Resize and insert data.
        src_end = lcl_ds[(eptr - new_rows) :]
        dest_end = gbl_ds[(eptr - new_rows) :]
        lcl_ds.resize(len(lcl_ds) + new_rows, axis=0)
        gbl_ds.resize(len(gbl_ds) + new_rows, axis=0)
        lcl_ds[iptr:eptr] = np.concatenate((lcl_ds[iptr : (eptr - new_rows)], lloc))
        lcl_ds[eptr:] = src_end
        gbl_ds[iptr:eptr] = np.concatenate((gbl_ds[iptr : (eptr - new_rows)], gloc))
        gbl_ds[eptr:] = dest_end

    def get_local_chunks(self, direction):
        with self._engine._read():
            with self._engine._handle("r") as handle:
                return [
                    Chunk.from_id(int(k), None)
                    for k in handle[self._path][direction].keys()
                ]

    def get_global_chunks(self, direction, local_):
        with self._engine._read():
            with self._engine._handle("r") as handle:
                return [
                    Chunk(k, None)
                    for k in handle[self._path][f"{direction}/{local_.id}"].attrs[
                        "chunk_list"
                    ]
                ]

    def nested_iter_connections(self, direction=None, local_=None, global_=None):
        return CSIterator(self, direction, local_, global_)

    def flat_iter_connections(self, direction=None, local_=None, global_=None):
        itr = CSIterator(self, direction, local_, global_)
        for dir in itr.get_dir_iter(direction):
            for lchunk in itr.get_local_iter(dir, local_):
                for gchunk in itr.get_global_iter(dir, lchunk, global_):
                    conns = self.load_connections(dir, lchunk, gchunk)
                    yield (dir, lchunk, gchunk, conns)

    def load_connections(self, direction, local_, global_, handle=None):
        if handle is None:
            with self._engine._read():
                with self._engine._handle("r") as handle:
                    return self._load_connections(direction, local_, global_, handle)
        else:
            return self._load_connections(direction, local_, global_, handle)

    def _load_connections(self, direction, local_, global_, handle):
        local_grp = handle[self._path][f"{direction}/{local_.id}"]
        start, end = self._get_insert_pointers(local_grp, global_)
        idx = slice(start, end)
        return (local_grp["local_locs"][idx], local_grp["global_locs"][idx])

    def load_local_connections(self, direction, local_, handle=None):
        if handle is None:
            with self._engine._read():
                with self._engine._handle("r") as handle:
                    return self._load_local(direction, local_, handle)
        else:
            return self._load_local(direction, local_, handle)

    def _load_local(self, direction, local_, handle):
        local_grp = handle[self._path][f"{direction}/{local_.id}"]
        global_locs = local_grp["global_locs"][()]
        chunks, ptrs = self._get_sorted_pointers(local_grp)
        col = np.repeat([c.id for c in chunks], np.diff(ptrs, append=len(global_locs)))
        return (local_grp["local_locs"][()], col, global_locs)


class CSIterator:
    def __init__(self, cs, direction=None, local_=None, global_=None):
        self._cs = cs
        self._dir = direction
        self._lchunks = local_
        self._gchunks = global_

    def __iter__(self):
        if self._dir is None:
            yield from (
                (dir, CSIterator(self._cs, dir, self._lchunks, self._gchunks))
                for dir in self.get_dir_iter(self._dir)
            )
        elif not isinstance(self._lchunks, Chunk):
            yield from (
                (lchunk, CSIterator(self._cs, self._dir, lchunk, self._gchunks))
                for lchunk in self.get_local_iter(self._dir, self._lchunks)
            )
        elif not isinstance(self._gchunks, Chunk):
            yield from (
                (gchunk, self._cs.load_connections(self._dir, self._lchunks, gchunk))
                for gchunk in self.get_global_iter(
                    self._dir, self._lchunks, self._gchunks
                )
            )
        else:
            yield self._cs.load_connections(self._dir, self._lchunks, self._gchunks)

    def get_dir_iter(self, dir):
        if dir is None:
            return ("inc", "out")
        else:
            return (dir,)

    def get_local_iter(self, dir, local_):
        if local_ is None:
            return self._cs.get_local_chunks(dir)
        elif isinstance(local_, Chunk):
            return (local_,)
        else:
            return iter(local_)

    def get_global_iter(self, dir, local_, global_):
        if global_ is None:
            return self._cs.get_global_chunks(dir, local_)
        elif isinstance(global_, Chunk):
            return (global_,)
        else:
            return iter(global_)


def _sort_triple(a, b):
    # Comparator for chunks by bitshift and sum of the coords.
    return (a[0] << 42 + a[1] << 21 + a[2]) > (b[0] << 42 + b[1] << 21 + b[2])



"""
The chunks module provides the tools for the HDF5 engine to store the chunked placement
data received from the placement module in seperate datasets to arbitrarily parallelize
and scale scaffold models.

The module provides the :class:`.ChunkLoader` mixin for
:class:`~.resource.Resource` objects (e.g. PlacementSet,
ConnectivitySet) to organize :class:`.ChunkedProperty` and :class:`.ChunkedCollection`
objects within them.
"""

from bsb.storage._chunks import Chunk
import numpy as np
import contextlib


class ChunkLoader:
    """
    :class:`~.resource.Resource` mixin to organize chunked properties
    and collections within itself.

    :param properties: An iterable of functions that construct :class:`.ChunkedProperty`.
    :type: Iterable
    :param properties: An iterable of names for constructing :class:`.ChunkedCollection`.
    :type: Iterable
    """

    def __init_subclass__(cls, properties=(), collections=(), **kwargs):
        super().__init_subclass__(**kwargs)
        cls._properties = list(properties)
        cls._collections = list(collections)

    def __init__(self):
        self._chunks = set()
        self._properties = []
        self._collections = []
        for prop_constr in self.__class__._properties:
            prop = prop_constr(self)
            self.__dict__[f"_{prop.name}_chunks"] = prop
            self._properties.append(prop)
        for col_name in self.__class__._collections:
            col = ChunkedCollection(self, col_name)
            self.__dict__[f"_{col}_chunks"] = col
            self._collections.append(col)

    def get_loaded_chunks(self):
        if not self._chunks:
            return self.get_all_chunks()
        else:
            return self._chunks.copy()

    def get_all_chunks(self):
        with self._engine._read():
            with self._engine._handle("r") as h:
                chunks = list(h[self._path + "/chunks"].keys())
                if chunks:
                    # If any chunks have been written, this HDF5 file is tagged with a
                    # chunk size
                    size = self._get_chunk_size(h)
        return [Chunk.from_id(int(c), size) for c in chunks]

    @contextlib.contextmanager
    def chunk_context(self, *chunks):
        old_chunks = self._chunks
        self._chunks = set(chunks)
        yield
        self._chunks = old_chunks

    def get_chunk_path(self, chunk=None):
        """
        Return the full HDF5 path of a chunk.

        :param chunk: Chunk
        :type chunk: :class:`bsb.storage.Chunk`
        :returns: HDF5 path
        :rtype: str
        """
        if chunk is None:
            return f"{self._path}/chunks/"
        else:
            return f"{self._path}/chunks/{chunk.id}"

    def load_chunk(self, chunk):
        """
        Add a chunk to read data from when loading properties/collections.
        """
        self._chunks.add(chunk if isinstance(chunk, Chunk) else Chunk(chunk, None))

    def unload_chunk(self, chunk):
        """
        Remove a chunk to read data from when loading properties/collections.
        """
        self._chunks.discard(chunk if isinstance(chunk, Chunk) else Chunk(chunk, None))

    def set_chunks(self, chunks):
        chunks = _to_chunklist(chunks)
        self._chunks = set(chunks)

    def clear_chunks(self):
        self._chunks = set()

    def require_chunk(self, chunk, handle=None):
        """
        Create a chunk if it doesn't exist yet, or do nothing.
        """
        if handle is not None:
            self._require(chunk, handle)
        else:
            with self._engine._write():
                with self._engine._handle("a") as handle:
                    self._require(chunk, handle)

    def _require(self, chunk, handle):
        path = self.get_chunk_path(chunk)
        if path in handle:
            return
        chunk_group = handle.create_group(path)
        self._set_chunk_size(handle, chunk.dimensions)
        for p in self._properties:
            chunk_group.create_dataset(
                f"{path}/{p.name}", p.shape, maxshape=p.maxshape, dtype=p.dtype
            )
        for c in self._collections:
            chunk_group.create_group(path + f"/{c.name}")

    def clear(self, chunks=None):
        if chunks is None:
            chunks = self.get_loaded_chunks()
        for chunk in chunks:
            for prop in self._properties:
                prop.clear(chunk)

    def _set_chunk_size(self, handle, size):
        fsize = handle.attrs.get("chunk_size", np.full(3, np.nan))
        if np.all(np.isnan(fsize)):
            handle.attrs["chunk_size"] = size
        elif not np.all(np.isnan(size)) and not np.allclose(fsize, size):
            raise Exception(f"Chunk size mismatch. File: {fsize}. Given: {size}")

    def _get_chunk_size(self, handle):
        return handle.attrs["chunk_size"]


def _to_chunklist(chunks):
    return [c if isinstance(c, Chunk) else Chunk(c, None) for c in chunks]


class ChunkedProperty:
    """
    Chunked properties are stored inside the ``chunks`` group of the :class:`.ChunkLoader`
    they belong to. Inside the ``chunks`` group another group is created per chunk, inside
    of which a dataset exists per property.
    """

    def __init__(self, loader, property, shape, dtype, insert=None, extract=None):
        self.loader = loader
        self.name = property
        self.dtype = dtype
        self.shape = shape
        self.insert = insert
        self.extract = extract
        maxshape = list(shape)
        maxshape[0] = None
        self.maxshape = tuple(maxshape)

    def load(self, raw=False):
        with self.loader._engine._read():
            if self.loader._chunks:
                chunks = self.loader._chunks
            else:
                chunks = self.loader.get_all_chunks()
            reader = self._chunk_reader(raw=raw)
            chunk_loader = map(reader, chunks)
            # Concatenate all non-empty chunks together
            chunked_data = tuple(c for c in chunk_loader if c.size)
        if not chunked_data:
            return np.empty(self.shape)
        return np.concatenate(chunked_data)

    def _chunk_reader(self, raw):
        """
        Create a chunk reader that either returns the raw data or extracts it.
        """

        def read_chunk(chunk):
            self.loader.require_chunk(chunk)
            with self.loader._engine._read():
                with self.loader._engine._handle("r") as f:
                    chunk_group = f[self.loader.get_chunk_path(chunk)]
                    if self.name not in chunk_group:
                        return np.empty(self.shape)
                    data = chunk_group[self.name][()]
                    return data

        # If this property has an extractor and we're not in raw mode, wrap the above
        # reader to extract the data
        if not (raw or self.extract is None):
            _f = read_chunk

            def read_chunk(chunk):
                data = self.extract(_f(chunk))
                # Allow only `np.ndarray`. Sorry things that quack, today we're checking
                # birth certificates. Purebred ducks only.
                if type(data) is not np.ndarray:
                    # Just kidding, as long as you quack you're welcome, but you'll have
                    # to change your family name.
                    data = np.array(data)
                return data

        # Return the created function
        return read_chunk

    def append(self, chunk, data):
        """
        Append data to a property chunk. Will create it if it doesn't exist.

        :param chunk: Chunk
        :type chunk: :class:`bsb.storage.Chunk`
        """
        if self.insert is not None:
            data = self.insert(data)
        with self.loader._engine._write():
            with self.loader._engine._handle("a") as f:
                self.loader.require_chunk(chunk, handle=f)
                chunk_group = f[self.loader.get_chunk_path(chunk)]
                if self.name not in chunk_group:
                    chunk_group.create_dataset(
                        self.name,
                        self.shape,
                        data=data,
                        maxshape=self.maxshape,
                        dtype=self.dtype,
                    )
                else:
                    dset = chunk_group[self.name]
                    start_pos = dset.shape[0]
                    dset.resize(start_pos + len(data), axis=0)
                    dset[start_pos:] = data

    def clear(self, chunk):
        with self.loader._engine._write():
            with self.loader._engine._handle("a") as f:
                self.loader.require_chunk(chunk, handle=f)
                chunk_group = f[self.loader.get_chunk_path(chunk)]
                if self.name not in chunk_group:
                    chunk_group.create_dataset(
                        self.name,
                        self.shape,
                        data=np.empty(self.shape),
                        maxshape=self.maxshape,
                        dtype=self.dtype,
                    )
                else:
                    dset = chunk_group[self.name]
                    dset.resize(0, axis=0)


class ChunkedCollection:
    """
    Chunked collections are stored inside the ``chunks`` group of the
    :class:`.ChunkLoader` they belong to. Inside the ``chunks`` group another group is
    created per chunk, inside of which a group exists per collection. Arbitrarily named
    datasets can be stored inside of this collection.
    """

    def __init__(self, loader, property):
        self.loader = loader
        self.name = property



from bsb.exceptions import (
    MissingMorphologyError,
    DatasetExistsError,
    DatasetNotFoundError,
)
from bsb.storage import Chunk
from bsb.storage.interfaces import PlacementSet as IPlacementSet
from bsb.morphologies import MorphologySet, RotationSet
from bsb.morphologies.selector import MorphologySelector
from .resource import Resource
from .chunks import ChunkLoader, ChunkedProperty
import numpy as np
import itertools


def _pos_prop(loader):
    ChunkedProperty(loader, "position", shape=(0, 3), dtype=float)


def _rot_prop(loader):
    ChunkedProperty(loader, "rotation", shape=(0, 3), dtype=float)


def _morpho_prop(loader):
    ChunkedProperty(loader, "morphology", shape=(0,), dtype=int)


_root = "/placement/"


class _MapSelector(MorphologySelector):
    def __new__(cls, *args, **kwargs):
        # Disable config node object creation
        return object.__new__(cls)

    def __init__(self, ps, names):
        self._ps = ps
        self._names = set(names)

    def validate(self, loaders):
        missing = set(self._names) - {m.get_meta()["name"] for m in loaders}
        if missing:
            raise MissingMorphologyError(
                "Morphology repository misses the following morphologies required by"
                + f" {self._ps.tag}: {', '.join(missing)}"
            )

    def pick(self, stored_morphology):
        name = stored_morphology.get_meta()["name"]
        return name in self._names


class PlacementSet(
    Resource,
    ChunkLoader,
    IPlacementSet,
    properties=(_pos_prop, _rot_prop, _morpho_prop),
    collections=("labels", "additional"),
):
    """
    Fetches placement data from storage.

    .. note::

        Use :meth:`Scaffold.get_placement_set <bsb.core.Scaffold.get_placement_set>` to
        correctly obtain a PlacementSet.
    """

    def __init__(self, engine, cell_type):
        tag = cell_type.name
        super().__init__(engine, _root + tag)
        IPlacementSet.__init__(self, engine, cell_type)
        ChunkLoader.__init__(self)
        if not self.exists(engine, cell_type):
            raise DatasetNotFoundError("PlacementSet '{}' does not exist".format(tag))

    @classmethod
    def create(cls, engine, cell_type):
        """
        Create the structure for this placement set in the HDF5 file. Placement sets are
        stored under ``/placement/<tag>``.
        """
        tag = cell_type.name
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                if path in h:
                    raise DatasetExistsError(f"PlacementSet '{tag}' already exists.")
                g = h.create_group(path)
                g.create_group("chunks")
        return cls(engine, cell_type)

    @staticmethod
    def exists(engine, cell_type):
        with engine._read():
            with engine._handle("r") as h:
                return "/placement/" + cell_type.name in h

    @classmethod
    def require(cls, engine, cell_type):
        tag = cell_type.name
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                g = h.require_group(path)
                g.require_group("chunks")
        return cls(engine, cell_type)

    def load_positions(self):
        """
        Load the cell positions.

        :raises: DatasetNotFoundError when there is no rotation information for this
           cell type.
        """
        try:
            return self._position_chunks.load()
        except DatasetNotFoundError:
            raise DatasetNotFoundError(
                "No position information for the '{}' placement set.".format(self.tag)
            )

    def load_rotations(self):
        """
        Load the cell rotations.

        :raises: DatasetNotFoundError when there is no rotation information for this
           cell type.
        """
        try:
            return RotationSet(self._rotation_chunks.load())
        except DatasetNotFoundError:
            raise DatasetNotFoundError(
                "No rotation information for the '{}' placement set.".format(self.tag)
            )

    def load_morphologies(self):
        """
        Load the cell morphologies.

        :raises: DatasetNotFoundError when the morphology data is not found.
        """
        try:
            return MorphologySet(
                self._get_morphology_loaders(), self._morphology_chunks.load()
            )
        except DatasetNotFoundError:
            raise DatasetNotFoundError(
                "No morphology information for the '{}' placement set.".format(self.tag)
            )

    def _get_morphology_loaders(self):
        loaded_names = set()
        stor_mor = []
        with self._engine._read():
            with self._engine._handle("r") as f:
                for chunk in self.get_loaded_chunks():
                    path = self.get_chunk_path(chunk)
                    _map = f[path].attrs.get("morphology_loaders", [])
                    cmlist = self._engine.morphologies.select(_MapSelector(self, _map))
                    stor_mor.extend(m for m in cmlist if m.name not in loaded_names)
                    loaded_names.update(m.name for m in cmlist)
        return stor_mor

    def _set_morphology_loaders(self, map):
        with self._engine._write():
            with self._engine._handle("a") as f:
                for chunk in self.get_loaded_chunks():
                    path = self.get_chunk_path(chunk)
                    f[path].attrs["morphology_loaders"] = map

    def __iter__(self):
        return itertools.zip_longest(
            self.load_positions(),
            self.load_morphologies(),
        )

    def __len__(self):
        return len(self._position_chunks.load())

    def append_data(
        self,
        chunk,
        positions=None,
        morphologies=None,
        rotations=None,
        additional=None,
        count=None,
    ):
        """
        Append data to the placement set.

        :param chunk: The chunk to store data in.
        :param positions: Cell positions
        :type positions: :class:`numpy.ndarray`
        :param rotations: Cell rotations
        :type rotations: ~bsb.morphologies.RotationSet
        :param morphologies: Cell morphologies
        :type morphologies: ~bsb.morphologies.MorphologySet
        :param count: Amount of entities to place. Excludes the use of any positional,
          rotational or morphological data.
        :type count: int
        """
        if not isinstance(chunk, Chunk):
            chunk = Chunk(chunk, None)
        if positions is not None:
            positions = np.array(positions, copy=False)
        if count is not None:
            if not (positions is None and morphologies is None):
                raise ValueError(
                    "The `count` keyword is reserved for creating entities,"
                    + " without any positional, or morphological data."
                )
            with self._engine._write():
                with self._engine._handle("a") as f:
                    self.require_chunk(chunk, handle=f)
                    path = self.get_chunk_path(chunk)
                    prev_count = f[path].attrs.get("entity_count", 0)
                    f[path].attrs["entity_count"] = prev_count + count

        if positions is not None:
            self._position_chunks.append(chunk, positions)
        if rotations is not None and morphologies is None:
            raise ValueError("Can't append rotations without morphologies.")
        if morphologies is not None:
            self._append_morphologies(chunk, morphologies)
            if rotations is None:
                rotations = np.zeros((len(morphologies), 3))
            self._rotation_chunks.append(chunk, rotations)

        if additional is not None:
            for key, ds in additional.items():
                self.append_additional(key, chunk, ds)

    def _append_morphologies(self, chunk, new_set):
        with self.chunk_context(chunk):
            morphology_set = self.load_morphologies().merge(new_set)
            self._set_morphology_loaders(morphology_set._serialize_loaders())
            self._morphology_chunks.clear(chunk)
            self._morphology_chunks.append(chunk, morphology_set.get_indices())

    def append_entities(self, chunk, count, additional=None):
        self.append_data(chunk, count=count, additional=additional)

    def append_additional(self, name, chunk, data):
        with self._engine._write():
            self.require_chunk(chunk)
            path = self.get_chunk_path(chunk) + "/additional/" + name
            with self._engine._handle("a") as f:
                if path not in f:
                    maxshape = list(data.shape)
                    maxshape[0] = None
                    f.create_dataset(path, data=data, maxshape=tuple(maxshape))
                else:
                    dset = f[path]
                    start_pos = dset.shape[0]
                    dset.resize(start_pos + len(data), axis=0)
                    dset[start_pos:] = data



from bsb.storage.interfaces import FileStore as IFileStore
from .resource import Resource
from bsb.exceptions import *
from uuid import uuid4
import json
import io

_root = "files"


class FileStore(Resource, IFileStore):
    def __init__(self, engine):
        super().__init__(engine, _root)

    def all(self):
        with self._engine._read():
            with self._engine._handle("r") as root:
                store = root[self._path]
                return {id: dict(f.attrs.items()) for id, f in store.items()}

    def load(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                ds = root[f"{self._path}/{id}"]
                return ds[()].decode("utf-8"), dict(ds.attrs)

    def stream(self, id, binary=False):
        content = self.load(id)
        if not binary:
            return io.TextIOWrapper(content)
        else:
            return io.BytesIO(content)

    def remove(self, id):
        with self._engine._write():
            with self._engine._handle("a") as root:
                del root[f"{self._path}/{id}"]

    def store(self, content, meta=None, id=None):
        if id is None:
            id = str(uuid4())
        if meta is None:
            meta = {}
        with self._engine._write():
            with self._engine._handle("a") as root:
                store = root[self._path]
                try:
                    ds = store.create_dataset(id, data=content)
                except ValueError:
                    raise Exception(f"File `{id}` already exists in store.")
                for k, v in meta.items():
                    ds.attrs[k] = v
        return id

    def load_active_config(self):
        from bsb.config import Configuration

        cfg_id = self._active_config_id()
        if cfg_id is None:
            raise Exception("No active config")
        else:
            content, meta = self.load(cfg_id)
            # It's a serialized Python dict, so it should be JSON readable. We don't use
            # evaluate because files might originate from untrusted sources.
            tree = json.loads(content)
            cfg = Configuration(**tree)
            cfg._meta = meta
            return cfg

    def store_active_config(self, config):
        id = self._active_config_id()
        self._engine.comm.barrier()
        if id is not None and self._engine.comm.get_rank() == 0:
            self.remove(id)
        config._meta["active_config"] = True
        active_id = None
        if self._engine.comm.get_rank() == 0:
            meta = {k: v for k, v in config._meta.items() if v is not None}
            active_id = self.store(json.dumps(config.__tree__()), meta)
        return self._engine.comm.bcast(active_id, root=0)

    def _active_config_id(self):
        match = (id for id, m in self.all().items() if m.get("active_config", False))
        return next(match, None)
